@@ -3,6 +3,7 @@ import pickle
 import logging
 import os
 import hashlib
+from shutil import move, rmtree
 import tempfile
 import inspect
 from fsspec import AbstractFileSystem, filesystem
@@ -105,11 +106,16 @@ class CachingFileSystem(AbstractFileSystem):
         self.compression = compression
         self.same_names = same_names
         self.target_protocol = (
-            target_protocol if isinstance(target_protocol, str) else fs.protocol
+            target_protocol
+            if isinstance(target_protocol, str)
+            else (fs.protocol if isinstance(fs.protocol, str) else fs.protocol[0])
         )
         self.load_cache()
         self.fs = fs if fs is not None else filesystem(target_protocol, **self.kwargs)
 
+    def _mkcache(self):
+        os.makedirs(self.storage[-1], exist_ok=True)
+
     def load_cache(self):
         """Read set of stored blocks from file"""
         cached_files = []
@@ -118,10 +124,14 @@ class CachingFileSystem(AbstractFileSystem):
             if os.path.exists(fn):
                 with open(fn, "rb") as f:
                     # TODO: consolidate blocks here
-                    cached_files.append(pickle.load(f))
+                    loaded_cached_files = pickle.load(f)
+                    for c in loaded_cached_files.values():
+                        if isinstance(c["blocks"], list):
+                            c["blocks"] = set(c["blocks"])
+                    cached_files.append(loaded_cached_files)
             else:
-                os.makedirs(storage, exist_ok=True)
                 cached_files.append({})
+        self._mkcache()
         self.cached_files = cached_files or [{}]
         self.last_cache = time.time()
 
@@ -154,10 +164,12 @@ class CachingFileSystem(AbstractFileSystem):
         fn2 = tempfile.mktemp()
         with open(fn2, "wb") as f:
             pickle.dump(cache, f)
-        os.replace(fn2, fn)
+        self._mkcache()
+        move(fn2, fn)
 
     def _check_cache(self):
         """Reload caches if time elapsed or any disappeared"""
+        self._mkcache()
         if not self.cache_check:
             # explicitly told not to bother checking
             return
@@ -168,14 +180,18 @@ class CachingFileSystem(AbstractFileSystem):
 
     def _check_file(self, path):
         """Is path in cache and still valid"""
+        path = self._strip_protocol(path)
+
         self._check_cache()
         if not path.startswith(self.target_protocol):
             store_path = self.target_protocol + "://" + path
             path = self.fs._strip_protocol(store_path)
+        else:
+            store_path = path
         for storage, cache in zip(self.storage, self.cached_files):
-            if path not in cache:
+            if store_path not in cache:
                 continue
-            detail = cache[path].copy()
+            detail = cache[store_path].copy()
             if self.check_files:
                 if detail["uid"] != self.fs.ukey(path):
                     continue
@@ -185,7 +201,42 @@ class CachingFileSystem(AbstractFileSystem):
             fn = os.path.join(storage, detail["fn"])
             if os.path.exists(fn):
                 return detail, fn
-        return False, None
+        return False
+
+    def clear_cache(self):
+        """Remove all files and metadat from the cache
+
+        In the case of multiple cache locations, this clears only the last one,
+        which is assumed to be the read/write one.
+        """
+        rmtree(self.storage[-1])
+        self.load_cache()
+
+    def pop_from_cache(self, path):
+        """Remove cached version of given file
+
+        Deletes local copy of the given (remote) path. If it is found in a cache
+        location which is not the last, it is assumed to be read-only, and
+        raises PermissionError
+        """
+        path = self._strip_protocol(path)
+        if not path.startswith(self.target_protocol):
+            store_path = self.target_protocol + "://" + path
+            path = self.fs._strip_protocol(store_path)
+        else:
+            store_path = path
+        _, fn = self._check_file(path)
+        if fn is None:
+            return
+        if fn.startswith(self.storage[-1]):
+            # is in in writable cache
+            os.remove(fn)
+            self.cached_files[-1].pop(store_path)
+            self.save_cache()
+        else:
+            raise PermissionError(
+                "Can only delete cached file in last, writable cache location"
+            )
 
     def _open(
         self,
@@ -224,9 +275,10 @@ class CachingFileSystem(AbstractFileSystem):
                 cache_options=cache_options,
                 **kwargs
             )
-        detail, fn = self._check_file(store_path)
+        detail = self._check_file(store_path)
         if detail:
             # file is in cache
+            detail, fn = detail
             hash, blocks = detail["fn"], detail["blocks"]
             if blocks is True:
                 # stored file is complete
@@ -248,6 +300,7 @@ class CachingFileSystem(AbstractFileSystem):
             logger.debug("Creating local sparse file for %s" % path)
 
         # call target filesystems open
+        self._mkcache()
         f = self.fs._open(
             path,
             mode=mode,
@@ -302,12 +355,20 @@ class CachingFileSystem(AbstractFileSystem):
             "__reduce__",
             "open",
             "cat",
+            "cat_file",
             "get",
             "read_block",
             "tail",
             "head",
             "_check_file",
             "_check_cache",
+            "clear_cache",
+            "pop_from_cache",
+            "_mkcache",
+            "local_file",
+            "_paths_from_path",
+            "open_many",
+            "commit_many",
         ]:
             # all the methods defined in this class. Note `open` here, since
             # it calls `_open`, but is actually in superclass
@@ -359,8 +420,60 @@ class WholeFileCacheFileSystem(CachingFileSystem):
     """
 
     protocol = "filecache"
+    local_file = True
 
-    def _open(self, path, mode="rb", **kwargs):
+    def open_many(self, open_files):
+        paths, store_paths = zip(*[self._paths_from_path(of.path) for of in open_files])
+        if "r" in open_files.mode:
+            self._mkcache()
+        else:
+            return [
+                LocalTempFile(self.fs, path, mode=open_files.mode, autocommit=False)
+                for path in paths
+            ]
+
+        if self.compression:
+            raise NotImplementedError
+        details = [self._check_file(sp) for sp in store_paths]
+        downpath = [p for p, d in zip(paths, details) if not d]
+        downstore = [p for p, d in zip(store_paths, details) if not d]
+        downfn0 = [
+            os.path.join(self.storage[-1], hash_name(p, self.same_names))
+            for p, d in zip(paths, details)
+        ]  # keep these path names for opening later
+        downfn = [fn for fn, d in zip(downfn0, details) if not d]
+        if downpath:
+            # skip if all files are already cached and up to date
+            self.fs.get(downpath, downfn)
+
+            # update metadata - only happens when downloads are successful
+            newdetail = [
+                {
+                    "fn": hash_name(path, self.same_names),
+                    "blocks": True,
+                    "time": time.time(),
+                    "uid": self.fs.ukey(path),
+                }
+                for path in downpath
+            ]
+            self.cached_files[-1].update(
+                {store_path: detail for store_path, detail in zip(downstore, newdetail)}
+            )
+            self.save_cache()
+
+        def firstpart(fn):
+            # helper to adapt both whole-file and simple-cache
+            return fn[1] if isinstance(fn, tuple) else fn
+
+        return [
+            open(firstpart(fn0) if fn0 else fn1, mode=open_files.mode)
+            for fn0, fn1 in zip(details, downfn0)
+        ]
+
+    def commit_many(self, open_files):
+        self.fs.put([f.fn for f in open_files], [f.path for f in open_files])
+
+    def _paths_from_path(self, path):
         path = self._strip_protocol(path)
 
         if not path.startswith(self.target_protocol):
@@ -368,10 +481,15 @@ class WholeFileCacheFileSystem(CachingFileSystem):
         else:
             store_path = path
         path = self.fs._strip_protocol(store_path)
+        return path, store_path
+
+    def _open(self, path, mode="rb", **kwargs):
+        path, store_path = self._paths_from_path(path)
         if "r" not in mode:
             return self.fs._open(path, mode=mode, **kwargs)
-        detail, fn = self._check_file(store_path)
+        detail = self._check_file(store_path)
         if detail:
+            detail, fn = detail
             hash, blocks = detail["fn"], detail["blocks"]
             if blocks is True:
                 logger.debug("Opening local copy of %s" % path)
@@ -384,10 +502,9 @@ class WholeFileCacheFileSystem(CachingFileSystem):
         else:
             hash = hash_name(path, self.same_names)
             fn = os.path.join(self.storage[-1], hash)
-            blocks = True
             detail = {
                 "fn": hash,
-                "blocks": blocks,
+                "blocks": True,
                 "time": time.time(),
                 "uid": self.fs.ukey(path),
             }
@@ -405,6 +522,7 @@ class WholeFileCacheFileSystem(CachingFileSystem):
                 else self.compression
             )
             f = compr[comp](f, mode="rb")
+        self._mkcache()
         with open(fn, "wb") as f2:
             if isinstance(f, AbstractBufferedFile):
                 # want no type of caching if just downloading whole thing
@@ -412,8 +530,9 @@ class WholeFileCacheFileSystem(CachingFileSystem):
             if getattr(f, "blocksize", 0) and f.size:
                 # opportunity to parallelise here
                 data = True
+                block = getattr(f, "blocksize", 5 * 2 ** 20)
                 while data:
-                    data = f.read(f.blocksize)
+                    data = f.read(block)
                     f2.write(data)
             else:
                 # this only applies to HTTP, should instead use streaming
@@ -422,7 +541,7 @@ class WholeFileCacheFileSystem(CachingFileSystem):
         return self._open(path, mode)
 
 
-class SimpleCacheFileSystem(CachingFileSystem):
+class SimpleCacheFileSystem(WholeFileCacheFileSystem):
     """Caches whole remote files on first access
 
     This class is intended as a layer over any other file system, and
@@ -431,9 +550,15 @@ class SimpleCacheFileSystem(CachingFileSystem):
     does not keep any metadata about the download time or file details.
     It is therefore safer to use in multi-threaded/concurrent situations.
 
+    This is the only of the caching filesystems that supports write: you will
+    be given a real local open file, and upon close and commit, it will be
+    uploaded to the target filesystem; the writability or the target URL is
+    not checked until that time.
+
     """
 
     protocol = "simplecache"
+    local_file = True
 
     def __init__(self, **kwargs):
         kw = kwargs.copy()
@@ -446,6 +571,7 @@ class SimpleCacheFileSystem(CachingFileSystem):
         self.cached_files = [{}]
 
     def _check_file(self, path):
+        self._check_cache()
         sha = hash_name(path, self.same_names)
         for storage in self.storage:
             fn = os.path.join(storage, sha)
@@ -467,7 +593,7 @@ class SimpleCacheFileSystem(CachingFileSystem):
             store_path = path
         path = self.fs._strip_protocol(store_path)
         if "r" not in mode:
-            return self.fs._open(path, mode=mode, **kwargs)
+            return LocalTempFile(self, path, mode=mode)
         fn = self._check_file(path)
         if fn:
             return open(fn, mode)
@@ -477,36 +603,73 @@ class SimpleCacheFileSystem(CachingFileSystem):
         logger.debug("Copying %s to local cache" % path)
         kwargs["mode"] = mode
 
-        with self.fs._open(path, **kwargs) as f, open(fn, "wb") as f2:
-            if isinstance(f, AbstractBufferedFile):
-                # want no type of caching if just downloading whole thing
-                f.cache = BaseCache(0, f.cache.fetcher, f.size)
-            if getattr(f, "blocksize", 0) and f.size:
-                # opportunity to parallelise here (if not compressed)
-                if self.compression:
-                    comp = (
-                        infer_compression(path)
-                        if self.compression == "infer"
-                        else self.compression
-                    )
-                    f = compr[comp](f, mode="rb")
+        self._mkcache()
+        if self.compression:
+            with self.fs._open(path, **kwargs) as f, open(fn, "wb") as f2:
+                if isinstance(f, AbstractBufferedFile):
+                    # want no type of caching if just downloading whole thing
+                    f.cache = BaseCache(0, f.cache.fetcher, f.size)
+                comp = (
+                    infer_compression(path)
+                    if self.compression == "infer"
+                    else self.compression
+                )
+                f = compr[comp](f, mode="rb")
                 data = True
                 while data:
-                    data = f.read(f.blocksize)
+                    block = getattr(f, "blocksize", 5 * 2 ** 20)
+                    data = f.read(block)
                     f2.write(data)
-            else:
-                # this only applies to HTTP, should instead use streaming
-                if self.compression:
-                    comp = (
-                        infer_compression(path)
-                        if self.compression == "infer"
-                        else self.compression
-                    )
-                    f = compr[comp](f, mode="rb")
-                f2.write(f.read())
+        else:
+            self.fs.get(path, fn)
         return self._open(path, mode)
 
 
+class LocalTempFile:
+    """A temporary local file, which will be uploaded on commit"""
+
+    def __init__(self, fs, path, fn=None, mode="wb", autocommit=True, seek=0):
+        fn = fn or tempfile.mktemp()
+        self.mode = mode
+        self.fn = fn
+        self.fh = open(fn, mode)
+        if seek:
+            self.fh.seek(seek)
+        self.path = path
+        self.fs = fs
+        self.closed = False
+        self.autocommit = autocommit
+
+    def __reduce__(self):
+        # always open in rb+ to allow continuing writing at a location
+        return (
+            LocalTempFile,
+            (self.fs, self.path, self.fn, "rb+", self.autocommit, self.tell()),
+        )
+
+    def __enter__(self):
+        return self.fh
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.close()
+
+    def close(self):
+        self.fh.close()
+        self.closed = True
+        if self.autocommit:
+            self.commit()
+
+    def discard(self):
+        self.fh.close()
+        os.remove(self.fn)
+
+    def commit(self):
+        self.fs.put(self.fn, self.path)
+
+    def __getattr__(self, item):
+        return getattr(self.fh, item)
+
+
 def hash_name(path, same_name):
     if same_name:
         hash = os.path.basename(path)
