@@ -44,8 +44,12 @@ class MemoryFileSystem(AbstractFileSystem):
                         "type": "file",
                     }
                 )
-            elif path and all(
-                (a == b) for a, b in zip(path.split("/"), p.strip("/").split("/"))
+            elif (
+                path
+                and len(path) < len(p.strip("/"))
+                and all(
+                    (a == b) for a, b in zip(path.split("/"), p.strip("/").split("/"))
+                )
             ):
                 # implicit directory
                 ppath = "/".join(p.split("/")[: len(path.split("/")) + 1])
@@ -80,15 +84,19 @@ class MemoryFileSystem(AbstractFileSystem):
             return out
         return sorted([f["name"] for f in out])
 
-    def mkdir(self, path):
+    def mkdir(self, path, create_parents=True, **kwargs):
         path = path.rstrip("/")
-        if path not in self.pseudo_dirs:
+        if create_parents and self._parent(path):
+            self.mkdir(self._parent(path), create_parents, **kwargs)
+        if self._parent(path) and not self.isdir(self._parent(path)):
+            raise NotADirectoryError(self._parent(path))
+        if path and path not in self.pseudo_dirs:
             self.pseudo_dirs.append(path)
 
     def rmdir(self, path):
         path = path.rstrip("/")
         if path in self.pseudo_dirs:
-            if self.ls(path) == []:
+            if not self.ls(path):
                 self.pseudo_dirs.remove(path)
             else:
                 raise OSError("Directory %s not empty" % path)
@@ -96,7 +104,7 @@ class MemoryFileSystem(AbstractFileSystem):
             raise FileNotFoundError(path)
 
     def exists(self, path):
-        return path in self.store
+        return path in self.store or path in self.pseudo_dirs
 
     def _open(
         self,
@@ -110,10 +118,12 @@ class MemoryFileSystem(AbstractFileSystem):
         if mode in ["rb", "ab", "rb+"]:
             if path in self.store:
                 f = self.store[path]
-                if mode == "rb":
-                    f.seek(0)
-                else:
+                if mode == "ab":
+                    # position at the end of file
                     f.seek(0, 2)
+                else:
+                    # position at the beginning of file
+                    f.seek(0)
                 return f
             else:
                 raise FileNotFoundError(path)
@@ -123,17 +133,28 @@ class MemoryFileSystem(AbstractFileSystem):
                 m.commit()
             return m
 
-    def copy(self, path1, path2, **kwargs):
-        self.store[path2] = MemoryFile(self, path2, self.store[path1].getbuffer())
+    def cp_file(self, path1, path2, **kwargs):
+        if self.isfile(path1):
+            self.store[path2] = MemoryFile(self, path2, self.store[path1].getbuffer())
+        elif self.isdir(path1):
+            if path2 not in self.pseudo_dirs:
+                self.pseudo_dirs.append(path2)
+        else:
+            raise FileNotFoundError
 
-    def cat(self, path):
+    def cat_file(self, path):
         try:
             return self.store[path].getvalue()
         except KeyError:
             raise FileNotFoundError(path)
 
     def _rm(self, path):
-        del self.store[path]
+        if self.isfile(path):
+            del self.store[path]
+        elif self.isdir(path):
+            self.rmdir(path)
+        else:
+            raise FileNotFoundError
 
     def size(self, path):
         """Size in bytes of the file at path"""
@@ -162,7 +183,9 @@ class MemoryFile(BytesIO):
         return self
 
     def close(self):
+        position = self.tell()
         self.size = self.seek(0, 2)
+        self.seek(position)
 
     def discard(self):
         pass
