@@ -4,11 +4,87 @@ from distutils.version import LooseVersion
 __all__ = ["registry", "get_filesystem_class", "default"]
 
 # mapping protocol: implementation class object
-registry = {}
+_registry = {}  # internal, mutable
+
+
+class ReadOnlyError(TypeError):
+    pass
+
+
+class ReadOnlyRegistry(dict):
+    """Dict-like registry, but immutable
+
+    Maps backend name to implementation class
+
+    To add backend implementations, use ``register_implementation``
+    """
+
+    def __init__(self, target):
+        self.target = target
+
+    def __getitem__(self, item):
+        return self.target[item]
+
+    def __delitem__(self, key):
+        raise ReadOnlyError
+
+    def __setitem__(self, key, value):
+        raise ReadOnlyError
+
+    def clear(self):
+        raise ReadOnlyError
+
+    def __contains__(self, item):
+        return item in self.target
+
+    def __iter__(self):
+        yield from self.target
+
+
+def register_implementation(name, cls, clobber=True, errtxt=None):
+    """Add implementation class to the registry
+
+    Parameters
+    ----------
+    name: str
+        Protocol name to associate with the class
+    cls: class or str
+        if a class: fsspec-compliant implementation class (normally inherits from
+        ``fsspec.AbstractFileSystem``, gets added straight to the registry. If a
+        str, the full path to an implementation class like package.module.class,
+        which gets added to known_implementations,
+        so the import is deferred until the filesystem is actually used.
+    clobber: bool (optional)
+        Whether to overwrite a protocol with the same name; if False, will raise
+        instead.
+    errtxt: str (optional)
+        If given, then a failure to import the given class will result in this
+        text being given.
+    """
+    if isinstance(cls, str):
+        if name in known_implementations and clobber is False:
+            raise ValueError(
+                "Name (%s) already in the known_implementations and clobber "
+                "is False" % name
+            )
+        known_implementations[name] = {
+            "class": cls,
+            "err": errtxt or "%s import failed for protocol %s" % (cls, name),
+        }
+
+    else:
+        if name in registry and clobber is False:
+            raise ValueError(
+                "Name (%s) already in the registry and clobber is False" % name
+            )
+        _registry[name] = cls
+
+
+registry = ReadOnlyRegistry(_registry)
 default = "file"
 
 # protocols mapped to the class which implements them. This dict can
-# be dynamically updated.
+# updated with register_implementation
 known_implementations = {
     "file": {"class": "fsspec.implementations.local.LocalFileSystem"},
     "memory": {"class": "fsspec.implementations.memory.MemoryFileSystem"},
@@ -21,11 +97,11 @@ known_implementations = {
     },
     "http": {
         "class": "fsspec.implementations.http.HTTPFileSystem",
-        "err": 'HTTPFileSystem requires "requests" to be installed',
+        "err": 'HTTPFileSystem requires "requests" and "aiohttp" to be installed',
     },
     "https": {
         "class": "fsspec.implementations.http.HTTPFileSystem",
-        "err": 'HTTPFileSystem requires "requests" to be installed',
+        "err": 'HTTPFileSystem requires "requests" and "aiohttp" to be installed',
     },
     "zip": {"class": "fsspec.implementations.zip.ZipFileSystem"},
     "gcs": {
@@ -36,6 +112,10 @@ known_implementations = {
         "class": "gcsfs.GCSFileSystem",
         "err": "Please install gcsfs to access Google Storage",
     },
+    "gdrive": {
+        "class": "gdrivefs.GoogleDriveFileSystem",
+        "err": "Please install gdrivefs for access to Google Drive",
+    },
     "sftp": {
         "class": "fsspec.implementations.sftp.SFTPFileSystem",
         "err": 'SFTPFileSystem requires "paramiko" to be installed',
@@ -62,6 +142,10 @@ known_implementations = {
         "class": "adlfs.AzureBlobFileSystem",
         "err": "Install adlfs to access Azure Datalake Gen2 and Azure Blob Storage",
     },
+    "az": {
+        "class": "adlfs.AzureBlobFileSystem",
+        "err": "Install adlfs to access Azure Datalake Gen2 and Azure Blob Storage",
+    },
     "cached": {"class": "fsspec.implementations.cached.CachingFileSystem"},
     "blockcache": {"class": "fsspec.implementations.cached.CachingFileSystem"},
     "filecache": {"class": "fsspec.implementations.cached.WholeFileCacheFileSystem"},
@@ -78,6 +162,10 @@ known_implementations = {
         "class": "fsspec.implementations.git.GitFileSystem",
         "err": "Install pygit2 to browse local git repos",
     },
+    "smb": {
+        "class": "fsspec.implementations.smb.SMBFileSystem",
+        "err": 'SMB requires "smbprotocol" or "smbprotocol[kerberos]" installed',
+    },
 }
 
 minversions = {"s3fs": LooseVersion("0.3.0"), "gcsfs": LooseVersion("0.3.0")}
@@ -102,7 +190,10 @@ def get_filesystem_class(protocol):
         if protocol not in known_implementations:
             raise ValueError("Protocol not known: %s" % protocol)
         bit = known_implementations[protocol]
-        registry[protocol] = _import_class(bit["class"])
+        try:
+            register_implementation(protocol, _import_class(bit["class"]))
+        except ImportError as e:
+            raise ImportError(bit["err"]) from e
     cls = registry[protocol]
     if getattr(cls, "protocol", None) in ("abstract", None):
         cls.protocol = protocol
